@@ -2,16 +2,16 @@ var indexSectionsWithContent =
 {
   0: ":_abcdefghijklmnopqrstuvwxyz~",
   1: "abcdefghijklmnopqrstuvw",
-  2: "opt",
+  2: "opst",
   3: "abcdefgilmnopqrstuv",
-  4: "abcdefghijklmnopqrstuvwxyz~",
-  5: "abcdfghilmnoprstvwxyz",
+  4: "_abcdefghijklmnopqrstuvwxyz~",
+  5: "abcdefghilmnoprstvwxyz",
   6: "_abcdefghilmnopqrstuvwxz",
-  7: "abcdfglmnrstv",
-  8: "abcdefghilmnoprstuvwxyz",
+  7: "abcdfgilmnrstv",
+  8: "abcdefghiklmnoprstuvwxyz",
   9: ":abcfilmoprstv",
   10: "ceopuv",
-  11: "cdfoprtu"
+  11: "bcdfortu"
 };
 
 var indexSectionNames =
